import torch


def create_rfft2_direction(shape: tuple, direction: tuple, direction_value=1j):
    """
    Create canonical rfft2 base vector in the spatial space
    This code is the functional implementation of the original main code in:
        https://github.com/LTS4/neural-anisotropy-directions/blob/master/directional_bias.py
    """
    v = torch.zeros(shape)  # Create empty vector
    v_fft = torch.fft.rfft2(v)
    v_fft[direction] = direction_value  # Select coordinate in fourier space
    v = torch.fft.irfft2(v_fft, s=shape[-2:])
    return v / v.norm()



import torch


class DirectionalLinearDataset(torch.utils.data.Dataset):
    """
    Creates a linearly separable dataset along the provided direction
    This code is a torch only rewrite of the original:
        https://github.com/LTS4/neural-anisotropy-directions/blob/master/directional_bias.py
    """
    def __init__(
            self,
            direction,
            shape,
            num_samples=10000,
            sigma=3,
            epsilon=1,
    ):
        self.direction = direction
        self.num_samples = num_samples
        self.sigma = sigma
        self.epsilon = epsilon
        self.shape = shape
        self.data, self.targets = self._generate_dataset(self.num_samples)
        super().__init__()

    def __getitem__(self, index):
        img, target = self.data[index], int(self.targets[index])

        return img, target

    def __len__(self):
        return self.num_samples

    def _generate_dataset(self, n_samples):
        if n_samples > 1:
            data_plus = self._generate_samples(n_samples // 2 + n_samples % 2, 0).type(torch.float32)
            labels_plus = torch.zeros([n_samples // 2 + n_samples % 2]).type(torch.long)
            data_minus = self._generate_samples(n_samples // 2, 1).type(torch.float32)
            labels_minus = torch.ones([n_samples // 2]).type(torch.long)
            data = torch.cat([data_plus, data_minus])
            labels = torch.cat([labels_plus, labels_minus])
        else:
            data = self._generate_samples(1, 0).type(torch.float32)
            labels = torch.zeros(1).type(torch.long)

        return data, labels

    def _generate_samples(self, n_samples, label):
        data = self._generate_noise_floor(n_samples)
        sign = 1 if label == 0 else -1
        data = sign * self.epsilon / 2 * self.direction[None, :] + self._project_orthogonal(data)
        return data

    def _generate_noise_floor(self, n_samples):
        data = self.sigma * torch.randn(n_samples, *self.shape)

        return data

    def _project(self, x):
        proj_x = torch.reshape(x, [x.shape[0], -1]) @ torch.reshape(self.direction, [-1, 1])
        return proj_x[:, :, None, None] * self.direction[None, :]

    def _project_orthogonal(self, x):
        return x - self._project(x)




from matplotlib import pyplot as plt
import typing as th
from .utils import make_subdirectories
import torch


def visualize_image(
        image: torch.Tensor,
        transform: th.Optional[th.Callable[[torch.Tensor], torch.Tensor]] = None,
        title: th.Optional[str] = None,
        no_axis: bool = True,
        cmap: str = 'BrBG',
        subplot: th.Optional[th.Tuple[int, int, int]] = None,
        path: th.Optional[str] = None
):
    """
    Visualize the given image (in the respective subplot if one was provided) with the provided settings. `transform`
    can be a function which transforms the image before visualizing it (e.g. fft2, etc.)
    """
    img = transform(image) if transform is not None else image
    if len(img.shape) == 3 and img.shape[0] == 3:
        img = img.permute(1, 2, 0)
    if len(img.shape) == 3 and img.shape[0] == 1:
        img = img[0]
    if subplot is not None:
        plt.subplot(subplot[0], subplot[1], subplot[2])
    value_max = max(img.max().abs().item(), img.min().abs().item())
    plt.imshow(img, cmap=cmap, vmin=-value_max, vmax=value_max)
    if title:
        plt.title(title)
    if no_axis:
        plt.axis('off')
    if path:
        make_subdirectories(path)
        plt.savefig(path)


def visualize_nads(
        nads: torch.Tensor,
        indices: th.Optional[th.Union[int, th.Tuple[int, int]]] = None,
        transform: th.Optional[th.Callable[[torch.Tensor], torch.Tensor]] = None,
        cmap: str = 'BrBG',
        columns_count: int = 5,
        title: th.Optional[str] = None,
        no_axis: bool = True,
        path: th.Optional[str] = None,
):
    """
    Visualize the given nad tensors (filter by indices if any where provided) with the provided settings. `transform`
    can be a function which transforms the image before visualizing it (e.g. fft2, etc.)
    Most of this code is based on the original
        https://github.com/LTS4/neural-anisotropy-directions/blob/master/nad_computation.py#L9
    """
    if indices is not None and isinstance(indices, int):
        visualize_image(
            nads[indices], transform=transform, cmap=cmap, no_axis=no_axis,
            title=f'NAD idx={indices}{" - " + title if title is not None else ""}'
        )
    else:
        indices = indices if indices is not None else (0, len(nads) + 1)
        count = indices[1] - indices[0]
        rows_count = count // columns_count

        plt.figure(figsize=(columns_count * 4, rows_count * 4))

        if title is not None:
            plt.suptitle(title)

        for i in range(*indices):
            visualize_image(
                nads[i].reshape(32, 32), transform=transform, cmap=cmap, no_axis=no_axis,
                title=f'NAD idx={i}', subplot=(rows_count, columns_count, i + 1))

        if path:
            make_subdirectories(path)
            plt.savefig(path)
        else:
            plt.show()


def visualize_spectrum(
        spectrum,
        x_label: th.Optional[str] = None,
        y_label: th.Optional[str] = None,
        title: th.Optional[str] = None,
        path: th.Optional[str] = None,
        plot_type: str = 'semilogy'
) -> None:
    """
    Visualize given spectrum with the given `plot_type`. The resulting graph will either be shown or saved to `path` in
    case one was provided.
    """
    getattr(plt, plot_type)(spectrum)
    plt.grid()
    if x_label:
        plt.xlabel(x_label)
    if y_label:
        plt.ylabel(y_label)
    plt.tight_layout()

    if title:
        plt.title(title)
    if path:
        make_subdirectories(path)
        plt.savefig(path)
    else:
        plt.show()



import os


def make_subdirectories(path: str, exist_ok: bool = True) -> None:
    """
    Create the subdirectories of the provided path and raise error if exists and `exist_ok` is false.
    """
    base_dir = '/'.join(path.split('/')[:-1])
    if base_dir:
        os.makedirs(base_dir, exist_ok=exist_ok)



import typing as th
import torch
from .jacobian import numerical_jacobian
from nads.nads import NADs


class GradientCovariance:
    """
    Gradient Covariance computation by MCMC sampling.
    Original code can be found in https://github.com/LTS4/neural-anisotropy-directions/blob/master/nad_computation.py#L9
    """
    def __init__(
            self,
            eval_point,
            model_cls: th.Type[type],
            model_params: th.Optional[dict] = None,
            device: str = 'cpu',
            force_eval=True,
            jacobian_fn: th.Callable[[th.Callable, torch.Tensor], torch.Tensor] = numerical_jacobian,
    ):
        """
        Gradient Covariance computation of NADs. `eval_points` is the initial input point for evaluating the expected
        w.r.t model parameters. `model_cls` should be a class variable for the desired model and `model_params` is an
        optional dictionary consisting the desired parameters to initialize the model with.`device` is where you want
        the computations to reside. if `force_eval` is set True
        """
        assert eval_point is not None, 'Evaluation point cannot be None'

        self.eval_point = eval_point.to(device)
        self.model_cls = model_cls
        self.model_params = model_params if model_params is not None else dict()
        self.device = device
        self.force_eval = force_eval
        self.jacobian_fn = jacobian_fn

    @property
    def sample_model(self):
        model = self.model_cls(**self.model_params).to(self.device)
        if self.force_eval:
            model = model.eval()
        return model

    @property
    def sample_grad(self):
        return self.jacobian_fn(self.sample_model, self.eval_point)

    def nads(self, num_samples=1000, grad_active=False):
        sample_gradient = torch.stack([self.sample_grad.view(-1) for i in range(num_samples)])
        with torch.set_grad_enabled(grad_active):
            u, s, vh = torch.linalg.svd(sample_gradient, full_matrices=False)
        return NADs(
            nads=vh.view(-1, *self.eval_point.shape)[:self.eval_point.numel()].detach(),
            eigen_values=s[:self.eval_point.numel()].detach(),
            name=self.model_cls.__name__,
            model_params=self.model_params,
        )



import torch


def numerical_jacobian(fn: torch.nn.Module, x: torch.Tensor, scale: float = 100) -> torch.Tensor:
    """
    Calculate the jacobian of the function w.r.t. the given input tensor x using numerical approximation.
    Original code can be found in https://github.com/LTS4/neural-anisotropy-directions/blob/master/utils.py#L82
    """
    shape = list(x.shape)
    n_dims = x.numel()
    v = torch.eye(n_dims, device=x.device).view([n_dims] + shape)
    jac = torch.zeros(n_dims, device=x.device)
    residual = 1 if n_dims % n_dims > 0 else 0
    with torch.no_grad():
        for n in range(n_dims // n_dims + residual):
            batch_plus = x[None, :] + scale * v[n * n_dims: (n + 1) * n_dims]
            batch_minus = x[None, :] - scale * v[n * n_dims: (n + 1) * n_dims]

            jac[n * n_dims: (n + 1) * n_dims] = ((fn(batch_plus) - fn(batch_minus)) / (2 * scale)).detach()[:, 0]
    return jac.view(shape)


def backprop_jacobian(fn: torch.nn.Module, x: torch.Tensor) -> torch.Tensor:
    """
    Calculate the jacobian of the function w.r.t. the given input tensor x using regular backprop.
    """
    x.requires_grad_()
    batch = x[None]
    return torch.autograd.grad(fn(batch), batch, allow_unused=True)[0][0]



from dataclasses import dataclass
from .visualization import visualize_spectrum, visualize_nads
from .utils import make_subdirectories
import typing as th
import torch


@dataclass
class NADs:
    """
    Utility dataclass which manages NADs saving/loading, visualization and potentially other operations
    """
    nads: torch.Tensor
    eigen_values: torch.Tensor
    name: th.Optional[str] = None
    model_params: th.Optional[dict] = None

    def to(self, device):
        assert self.nads is not None and self.eigen_values is not None, \
            'neither `nads` nor `eigen_velues` should be None'
        self.nads = self.nads.to(device)
        self.eigen_values = self.eigen_values.to(device)

    def save(self, path):
        """
        Save nads & eigen values along with the name of the model and its initialization parameters to the provided
        path. Subdirectories are automatically created.
        """
        make_subdirectories(path)
        self.to('cpu')
        torch.save(
            dict(nads=self.nads, eigen_values=self.eigen_values, name=self.name, model_params=self.model_params), path)

    def __getitem__(self, item):
        return self.nads.__getitem__(item)

    @classmethod
    def load(cls, path):
        """
        Load a NADs object from the provided path.
        """
        return NADs(**torch.load(path))

    def visualize_spectrum(self, **kwargs):
        """
        Visualize the NAD eigen_values spectrum. Check `visualization.visualize_spectrum` to see how keyword arguments
        affect the results.
        """
        kwargs['title'] = kwargs.get('title', f'{self.name} NADs eigen spectrum' if self.name else None)
        kwargs['x_label'] = kwargs.get('x_label', 'NAD Index')
        kwargs['y_label'] = kwargs.get('y_label', '$\sigma$')
        visualize_spectrum(self.eigen_values.tolist(), **kwargs)

    def visualize_nads(self, **kwargs):
        """
        Visualize the NAD tesnor images. Check `visualization.visualize_nads` to see how keyword arguments affect the
        results.
        """
        kwargs['title'] = kwargs.get('title', f'{self.name} NADs' if self.name else None)
        kwargs['indices'] = kwargs.get('indices', (0, 10))
        visualize_nads(self.nads, **kwargs)



from setuptools import setup, find_packages

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='nads',
    packages=find_packages(exclude=['examples', 'docs']),  # add other exclusions in future
    version='0.0.2',
    license='MIT',
    description='Neural Anisotropy Directions Toolbox',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Vahid Zehtab',
    author_email='vahid98zee@gmail.com',
    url='https://github.com/vahidzee/nads',
    keywords=[
        'artificial intelligence',
        'inductive bias',
        'deep learning'
    ],
    install_requires=[
        'torch>=1.9',
        'matplotlib>=3.3',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
    ],
)
